From 0ecc9324ef2837019e63081247a954f5a983ecc3 Mon Sep 17 00:00:00 2001
From: =?utf-8?q?Etienne=20B=C3=B6mcke?= <etienne.bomcke@uclouvain.be>
Date: Mon, 23 Jun 2008 15:22:23 +0200
Subject: [PATCH] Add multisliced encoding support

---
 common/common.c   |    5 +++-
 common/common.h   |    6 ++--
 encoder/analyse.c |   35 ++++++++++++++++++-------
 encoder/encoder.c |   73 ++++++++++++++++++++++++++++++++++++++++-------------
 x264.c            |   18 +++++++++++--
 x264.h            |    7 +++++
 6 files changed, 109 insertions(+), 35 deletions(-)

diff --git a/common/common.c b/common/common.c
index 8a8f660..52eaa7e 100644
--- a/common/common.c
+++ b/common/common.c
@@ -62,6 +62,9 @@ void    x264_param_default( x264_param_t *param )
     param->i_fps_num       = 25;
     param->i_fps_den       = 1;
     param->i_level_idc     = 51; /* as close to "unrestricted" as we can get */
+    param->i_slice_size_threshold = 0;  /* by default: slice size threshold is infinity */
+    param->i_slice_size_mb_threshold = 0;  /* by default: slice size threshold is infinity */
+    param->i_slice_qty = 1;  /* by default: frames are encoded in one slice */
 
     /* Encoder parameters */
     param->i_frame_reference = 1;
@@ -759,7 +762,7 @@ void x264_free( void *p )
  ****************************************************************************/
 void *x264_realloc( void *p, int i_size )
 {
-#ifdef HAVE_MALLOC_H
+#ifdef HAVE_MALLOC_H || defined(SYS_MACOSX)
     return realloc( p, i_size );
 #else
     int       i_old_size = 0;
diff --git a/common/common.h b/common/common.h
index 0636394..aa5f0fe 100644
--- a/common/common.h
+++ b/common/common.h
@@ -48,8 +48,7 @@
 
 #define X264_BFRAME_MAX 16
 #define X264_THREAD_MAX 128
-#define X264_SLICE_MAX 4
-#define X264_NAL_MAX (4 + X264_SLICE_MAX)
+#define X264_INIT_NAL_NUM 8 
 
 // number of pixels (per thread) in progress at any given time.
 // 16 for the macroblock in progress + 3 for deblocking + 3 for motion compensation filter + 2 for extra safety
@@ -254,7 +253,8 @@ struct x264_t
     struct
     {
         int         i_nal;
-        x264_nal_t  nal[X264_NAL_MAX];
+        int         i_nals_allocated;
+        x264_nal_t  *nal;
         int         i_bitstream;    /* size of p_bitstream */
         uint8_t     *p_bitstream;   /* will hold data for all nal */
         bs_t        bs;
diff --git a/encoder/analyse.c b/encoder/analyse.c
index 9200ace..c3e9379 100644
--- a/encoder/analyse.c
+++ b/encoder/analyse.c
@@ -242,6 +242,7 @@ static void x264_mb_analyse_init( x264_t *h, x264_mb_analysis_t *a, int i_qp )
 
         /* Calculate max allowed MV range */
 #define CLIP_FMV(mv) x264_clip3( mv, -i_fmv_range, i_fmv_range-1 )
+
         h->mb.mv_min[0] = 4*( -16*h->mb.i_mb_x - 24 );
         h->mb.mv_max[0] = 4*( 16*( h->sps->i_mb_width - h->mb.i_mb_x - 1 ) + 24 );
         h->mb.mv_min_spel[0] = CLIP_FMV( h->mb.mv_min[0] );
@@ -352,23 +353,30 @@ static void x264_mb_analyse_init( x264_t *h, x264_mb_analysis_t *a, int i_qp )
 /* Max = 4 */
 static void predict_16x16_mode_available( unsigned int i_neighbour, int *mode, int *pi_count )
 {
-    if( i_neighbour & MB_TOPLEFT )
+    int b_l = i_neighbour & MB_LEFT;
+    int b_t = i_neighbour & MB_TOP;
+    if( b_l && b_t )
     {
         /* top and left available */
         *mode++ = I_PRED_16x16_V;
         *mode++ = I_PRED_16x16_H;
         *mode++ = I_PRED_16x16_DC;
-        *mode++ = I_PRED_16x16_P;
-        *pi_count = 4;
+        *pi_count = 3;
+        if( i_neighbour & MB_TOPLEFT )
+        {
+            /* top left available*/
+	    *mode++ = I_PRED_16x16_P;
+	    *pi_count = 4;
+        }
     }
-    else if( i_neighbour & MB_LEFT )
+    else if( b_l )
     {
         /* left available*/
         *mode++ = I_PRED_16x16_DC_LEFT;
         *mode++ = I_PRED_16x16_H;
         *pi_count = 2;
     }
-    else if( i_neighbour & MB_TOP )
+    else if( b_t )
     {
         /* top available*/
         *mode++ = I_PRED_16x16_DC_TOP;
@@ -386,23 +394,30 @@ static void predict_16x16_mode_available( unsigned int i_neighbour, int *mode, i
 /* Max = 4 */
 static void predict_8x8chroma_mode_available( unsigned int i_neighbour, int *mode, int *pi_count )
 {
-    if( i_neighbour & MB_TOPLEFT )
+    int b_l = i_neighbour & MB_LEFT;
+    int b_t = i_neighbour & MB_TOP;
+    if( b_l && b_t )
     {
         /* top and left available */
         *mode++ = I_PRED_CHROMA_V;
         *mode++ = I_PRED_CHROMA_H;
         *mode++ = I_PRED_CHROMA_DC;
-        *mode++ = I_PRED_CHROMA_P;
-        *pi_count = 4;
+        *pi_count = 3;
+        if( i_neighbour & MB_TOPLEFT )
+        {
+            /* top left available */
+	    *mode++ = I_PRED_CHROMA_P;
+	    *pi_count = 4;
+        }
     }
-    else if( i_neighbour & MB_LEFT )
+    else if( b_l )
     {
         /* left available*/
         *mode++ = I_PRED_CHROMA_DC_LEFT;
         *mode++ = I_PRED_CHROMA_H;
         *pi_count = 2;
     }
-    else if( i_neighbour & MB_TOP )
+    else if( b_t )
     {
         /* top available*/
         *mode++ = I_PRED_CHROMA_DC_TOP;
diff --git a/encoder/encoder.c b/encoder/encoder.c
index f2710ab..fdd2e9f 100644
--- a/encoder/encoder.c
+++ b/encoder/encoder.c
@@ -85,7 +85,7 @@ static void x264_slice_header_init( x264_t *h, x264_slice_header_t *sh,
     sh->pps = pps;
 
     sh->i_first_mb  = 0;
-    sh->i_last_mb   = h->sps->i_mb_width * h->sps->i_mb_height;
+    sh->i_last_mb   = (param->i_slice_size_mb_threshold == 0) ? h->sps->i_mb_width * h->sps->i_mb_height - 1 : param->i_slice_size_mb_threshold - 1;
     sh->i_pps_id    = pps->i_id;
 
     sh->i_frame_num = i_frame;
@@ -517,6 +517,11 @@ static int x264_validate_parameters( x264_t *h )
         h->param.analyse.b_ssim = 0;
     }
 
+    /* Set the slice size in MB according to the requested slice number */
+    if (h->param.i_slice_qty > 0)
+        h->param.i_slice_size_mb_threshold = ceil(((double)(h->param.i_width/16 * h->param.i_height/16))/h->param.i_slice_qty);
+
+
     /* ensure the booleans are 0 or 1 so they can be used in math */
 #define BOOLIFY(x) h->param.x = !!h->param.x
     BOOLIFY( b_cabac );
@@ -699,6 +704,8 @@ x264_t *x264_encoder_open   ( x264_param_t *param )
             *h->thread[i] = *h;
         h->thread[i]->fdec = x264_frame_pop_unused( h );
         h->thread[i]->out.p_bitstream = x264_malloc( h->out.i_bitstream );
+        h->thread[i]->out.nal = x264_malloc( sizeof ( x264_nal_t) * X264_INIT_NAL_NUM);
+        h->thread[i]->out.i_nals_allocated = X264_INIT_NAL_NUM;
         if( x264_macroblock_cache_init( h->thread[i] ) < 0 )
             return NULL;
     }
@@ -770,6 +777,7 @@ static void x264_nal_start( x264_t *h, int i_type, int i_ref_idc )
     nal->i_ref_idc = i_ref_idc;
     nal->i_type    = i_type;
 
+    nal->i_numb_mbs= 0;
     nal->i_payload= 0;
     nal->p_payload= &h->out.p_bitstream[bs_pos( &h->out.bs ) / 8];
 }
@@ -778,6 +786,13 @@ static void x264_nal_end( x264_t *h )
     x264_nal_t *nal = &h->out.nal[h->out.i_nal];
     nal->i_payload = &h->out.p_bitstream[bs_pos( &h->out.bs ) / 8] - nal->p_payload;
     h->out.i_nal++;
+    //if number of allocated nals is not enough, re-allocate a larger one (steps of 1) 
+    if(h->out.i_nal >= h->out.i_nals_allocated)
+    {
+        h->out.nal = x264_realloc(h->out.nal, sizeof ( x264_nal_t) * (h->out.i_nals_allocated + 1));
+//        h->out.nal = realloc(h->out.nal, sizeof ( x264_nal_t) * (h->out.i_nals_allocated + 1));
+        h->out.i_nals_allocated++;
+    }
 }
 
 /****************************************************************************
@@ -1046,9 +1061,7 @@ static void x264_slice_write( x264_t *h )
     int i_skip;
     int mb_xy, i_mb_x, i_mb_y;
     int i;
-
-    /* init stats */
-    memset( &h->stat.frame, 0, sizeof(h->stat.frame) );
+    int slice_bits_remain;
 
     /* Slice */
     x264_nal_start( h, h->i_nal_type, h->i_nal_ref_idc );
@@ -1066,12 +1079,15 @@ static void x264_slice_write( x264_t *h )
     }
     h->mb.i_last_qp = h->sh.i_qp;
     h->mb.i_last_dqp = 0;
+    
+    /* slice size threshold in bits */
+    slice_bits_remain = (h->param.i_slice_size_threshold<<3);
 
     i_mb_y = h->sh.i_first_mb / h->sps->i_mb_width;
     i_mb_x = h->sh.i_first_mb % h->sps->i_mb_width;
     i_skip = 0;
 
-    while( (mb_xy = i_mb_x + i_mb_y * h->sps->i_mb_width) < h->sh.i_last_mb )
+    while( (mb_xy = i_mb_x + i_mb_y * h->sps->i_mb_width) <= h->sh.i_last_mb )
     {
         int mb_spos = bs_pos(&h->out.bs) + x264_cabac_pos(&h->cabac);
 
@@ -1166,6 +1182,13 @@ static void x264_slice_write( x264_t *h )
             i_mb_y++;
             i_mb_x = 0;
         }
+        //end slice if a) finished frame b) there is a max slice size and it has been exceeded
+        slice_bits_remain -= (bs_pos(&h->out.bs) + x264_cabac_pos(&h->cabac) - mb_spos);
+        if ((mb_xy > h->sh.i_last_mb)
+            || ((h->param.i_slice_size_threshold > 0) && (slice_bits_remain<0))) 
+        {
+           break;
+        }
     }
 
     if( h->param.b_cabac )
@@ -1181,16 +1204,17 @@ static void x264_slice_write( x264_t *h )
         bs_rbsp_trailing( &h->out.bs );
     }
 
+    h->out.nal[h->out.i_nal].i_numb_mbs = mb_xy - h->sh.i_first_mb;
     x264_nal_end( h );
 
-    x264_fdec_filter_row( h, h->sps->i_mb_height );
-
-    /* Compute misc bits */
-    h->stat.frame.i_misc_bits = bs_pos( &h->out.bs )
-                              + NALU_OVERHEAD * 8
-                              - h->stat.frame.i_itex_bits
-                              - h->stat.frame.i_ptex_bits
-                              - h->stat.frame.i_hdr_bits;
+    h->sh.i_first_mb = mb_xy;
+	
+	if( h->param.i_slice_size_mb_threshold > 0 )
+	{
+		int i_frame_size = h->sps->i_mb_width * h->sps->i_mb_height;
+		int i_last_mb = mb_xy + h->param.i_slice_size_mb_threshold - 1;
+		h->sh.i_last_mb = (i_last_mb > i_frame_size - 1) ? i_frame_size - 1 : i_last_mb;
+	}
 }
 
 static void x264_thread_sync_context( x264_t *dst, x264_t *src )
@@ -1222,15 +1246,27 @@ static void x264_thread_sync_stat( x264_t *dst, x264_t *src )
 
 static int x264_slices_write( x264_t *h )
 {
-    int i_frame_size;
+    int i_frame_size = 0;
 
 #if VISUALIZE
     if( h->param.b_visualize )
         x264_visualize_init( h );
 #endif
 
-    x264_stack_align( x264_slice_write, h );
-    i_frame_size = h->out.nal[h->out.i_nal-1].i_payload;
+    /* init stats */
+    memset( &h->stat.frame, 0, sizeof(h->stat.frame) );
+    while(h->sh.i_first_mb < h->sh.i_last_mb)
+    {
+        x264_stack_align( x264_slice_write, h );
+        i_frame_size += h->out.nal[h->out.i_nal-1].i_payload;
+    }
+    h->stat.frame.i_misc_bits = bs_pos( &h->out.bs )
+                              + (h->out.i_nal*NALU_OVERHEAD * 8)
+                              - h->stat.frame.i_itex_bits
+                              - h->stat.frame.i_ptex_bits
+                              - h->stat.frame.i_hdr_bits;
+
+    x264_fdec_filter_row( h, h->sps->i_mb_height );
 
 #if VISUALIZE
     if( h->param.b_visualize )
@@ -1653,9 +1689,9 @@ static void x264_encoder_frame_end( x264_t *h, x264_t *thread_current,
 
     /* Slice stat */
     h->stat.i_slice_count[h->sh.i_type]++;
-    h->stat.i_slice_size[h->sh.i_type] += h->out.i_frame_size + NALU_OVERHEAD;
+    h->stat.i_slice_size[h->sh.i_type] += (h->out.i_frame_size + (h->out.i_nal*NALU_OVERHEAD));
     h->stat.f_slice_qp[h->sh.i_type] += h->fdec->f_qp_avg_aq;
-
+   
     for( i = 0; i < X264_MBTYPE_MAX; i++ )
         h->stat.i_mb_count[h->sh.i_type][i] += h->stat.frame.i_mb_count[i];
     for( i = 0; i < 2; i++ )
@@ -1948,6 +1984,7 @@ void    x264_encoder_close  ( x264_t *h )
     {
         x264_macroblock_cache_end( h->thread[i] );
         x264_free( h->thread[i]->out.p_bitstream );
+        x264_free( h->thread[i]->out.nal);
         x264_free( h->thread[i] );
     }
 }
diff --git a/x264.c b/x264.c
index d78461f..32cf602 100644
--- a/x264.c
+++ b/x264.c
@@ -171,6 +171,8 @@ static void Help( x264_param_t *defaults, int b_longhelp )
     H1( "      --no-deblock            Disable loop filter\n" );
     H0( "  -f, --deblock <alpha:beta>  Loop filter AlphaC0 and Beta parameters [%d:%d]\n",
                                        defaults->i_deblocking_filter_alphac0, defaults->i_deblocking_filter_beta );
+    H0( "      --max-slicesize <integer>  Limit the size of each slice (in bytes) (0 => unlimited)\n");
+    H0( "      --slices <integer>      Set the number of slices used (default : 1) \n");
     H0( "      --interlaced            Enable pure-interlaced mode\n" );
     H0( "\n" );
     H0( "Ratecontrol:\n" );
@@ -367,6 +369,8 @@ static int  Parse( int argc, char **argv,
 #define OPT_PROGRESS 261
 #define OPT_VISUALIZE 262
 #define OPT_LONGHELP 263
+#define OPT_SLICE_MAX 264
+#define OPT_SLICE_QTY 265
 
         static struct option long_options[] =
         {
@@ -440,6 +444,8 @@ static int  Parse( int argc, char **argv,
             { "zones",   required_argument, NULL, 0 },
             { "qpfile",  required_argument, NULL, OPT_QPFILE },
             { "threads", required_argument, NULL, 0 },
+            { "max-slicesize", required_argument, NULL, OPT_SLICE_MAX },
+            { "slices", required_argument, NULL, OPT_SLICE_QTY },
             { "thread-input", no_argument,  NULL, OPT_THREAD_INPUT },
             { "non-deterministic", no_argument, NULL, 0 },
             { "no-psnr", no_argument,       NULL, 0 },
@@ -543,6 +549,12 @@ static int  Parse( int argc, char **argv,
                 param->i_scenecut_threshold = -1;
                 param->b_bframe_adaptive = 0;
                 break;
+            case OPT_SLICE_MAX:
+                param->i_slice_size_threshold = atoi(optarg);
+                break;
+            case OPT_SLICE_QTY:
+                param->i_slice_qty = atoi(optarg);
+                break;
             case OPT_THREAD_INPUT:
                 b_thread_input = 1;
                 break;
@@ -732,10 +744,9 @@ static int  Encode_frame( x264_t *h, hnd_t hout, x264_picture_t *pic )
     {
         fprintf( stderr, "x264 [error]: x264_encoder_encode failed\n" );
     }
-
     for( i = 0; i < i_nal; i++ )
     {
-        int i_size;
+		int i_size;
 
         if( mux_buffer_size < nal[i].i_payload * 3/2 + 4 )
         {
@@ -745,9 +756,10 @@ static int  Encode_frame( x264_t *h, hnd_t hout, x264_picture_t *pic )
         }
 
         i_size = mux_buffer_size;
-        x264_nal_encode( mux_buffer, &i_size, 1, &nal[i] );
+		x264_nal_encode( mux_buffer, &i_size, 1, &nal[i] );
         i_file += p_write_nalu( hout, mux_buffer, i_size );
     }
+
     if (i_nal)
         p_set_eop( hout, &pic_out );
 
diff --git a/x264.h b/x264.h
index 0e257a1..3acca0d 100644
--- a/x264.h
+++ b/x264.h
@@ -288,6 +288,12 @@ typedef struct x264_param_t
     int b_aud;                  /* generate access unit delimiters */
     int b_repeat_headers;       /* put SPS/PPS before each keyframe */
     int i_sps_id;               /* SPS and PPS id number */
+	
+	/* Slice related parameters */
+    int i_slice_size_threshold;   /* start new slice if current slice size exceeds this threshold in bytes. ( 0 is interpreted as infinity)   
+                                 * Note: provides no guarantee for maximum size of slice. Worst case: threshold + maximum bytes possible to encode one mb */
+	int i_slice_size_mb_threshold; /* start new slice if current slice size exceeds this threshold in #MB. ( 0 is interpreted as infinity) */
+    int i_slice_qty;
 } x264_param_t;
 
 typedef struct {
@@ -388,6 +394,7 @@ typedef struct
 {
     int i_ref_idc;  /* nal_priority_e */
     int i_type;     /* nal_unit_type_e */
+    int i_numb_mbs; /* number of macroblocks in this NAL */
 
     /* This data are raw payload */
     int     i_payload;
-- 
1.5.4

